@@ -1,3 +1,6 @@
+import abc
+from warnings import warn
+
 from lbmpy.advanced_streaming.utility import AccessPdfValues, Timestep
 from pystencils.simp.assignment_collection import AssignmentCollection
 from pystencils import Assignment, Field
@@ -5,13 +8,13 @@ from lbmpy.boundaries.boundaryhandling import LbmWeightInfo
 from pystencils.data_types import create_type
 from pystencils.sympyextensions import get_symmetric_part
 from lbmpy.simplificationfactory import create_simplification_strategy
-from lbmpy.advanced_streaming.indexing import NeighbourOffsetArrays
+from lbmpy.advanced_streaming.indexing import NeighbourOffsetArrays, MirroredStencilDirections
 from pystencils.stencil import offset_to_direction_string, direction_string_to_offset, inverse_direction
 
 import sympy as sp
 
 
-class LbBoundary:
+class LbBoundary(abc.ABC):
     """Base class that all boundaries should derive from.
 
     Args:
@@ -79,14 +82,22 @@ class LbBoundary:
     def name(self, new_value):
         self._name = new_value
 
+    def __hash__(self):
+        return hash(self.name)
+
+    def __eq__(self, other):
+        if not isinstance(other, type(self)):
+            return False
+        return self.__dict__ == other.__dict__
+
 
 # end class Boundary
 
 
 class NoSlip(LbBoundary):
     """
-        No-Slip, (half-way) simple bounce back boundary condition, enforcing zero velocity at obstacle.
-        Extended for use with any streaming pattern.
+    No-Slip, (half-way) simple bounce back boundary condition, enforcing zero velocity at obstacle.
+    Extended for use with any streaming pattern.
 
     Args:
         name: optional name of the boundary.
@@ -99,17 +110,126 @@ class NoSlip(LbBoundary):
     def __call__(self, f_out, f_in, dir_symbol, inv_dir, lb_method, index_field):
         return Assignment(f_in(inv_dir[dir_symbol]), f_out(dir_symbol))
 
-    def __hash__(self):
-        return hash(self.name)
-
-    def __eq__(self, other):
-        if not isinstance(other, NoSlip):
-            return False
-        return self.name == other.name
-
 
 # end class NoSlip
 
+class FreeSlip(LbBoundary):
+    """
+    Free-Slip boundary condition, which enforces a zero normal fluid velocity :math:`u_n = 0` but places no restrictions
+    on the tangential fluid velocity :math:`u_t`.
+
+    Args:
+        stencil: LBM stencil which is used for the simulation
+        normal_direction: optional normal direction. If the Free slip boundary is applied to a certain side in the
+                          domain it is not necessary to calculate the normal direction since it can be stated for all
+                          boundary cells. This reduces the memory space for the index array significantly.
+        name: optional name of the boundary.
+    """
+
+    def __init__(self, stencil, normal_direction=None, name=None):
+        """Set an optional name here, to mark boundaries, for example for force evaluations"""
+        self.stencil = stencil
+
+        if normal_direction and len(normal_direction) - normal_direction.count(0) != 1:
+            raise ValueError("It is only possible to pre specify the normal direction for simple situations."
+                             "This means if the free slip boundary is applied to a straight wall or side in the "
+                             "simulation domain. A possible value for example would be (0, 1, 0) if the "
+                             "free slip boundary is applied to the northern wall. For more complex situations "
+                             "the normal direction has to be calculated for each cell. This is done when "
+                             "the normal direction is not defined for this class")
+
+        if normal_direction:
+            self.mirror_axis = normal_direction.index(*[dir for dir in normal_direction if dir != 0])
+
+        self.normal_direction = normal_direction
+        self.dim = len(stencil[0])
+
+        if name is None and normal_direction:
+            name = f"Free Slip : {offset_to_direction_string([-x for x in normal_direction])}"
+
+        super(FreeSlip, self).__init__(name)
+
+    def init_callback(self, boundary_data, **_):
+        if len(boundary_data.index_array) > 1e6:
+            warn(f"The calculation of the normal direction for each cell might take a long time, because "
+                 f"{len(boundary_data.index_array)} cells are marked as Free Slip boundary cells. Consider specifying "
+                 f" the normal direction beforehand, which is possible if it is equal for all cells (e.g. at a wall)")
+
+        dim = boundary_data.dim
+        coords = [coord for coord, _ in zip(['x', 'y', 'z'], range(dim))]
+
+        boundary_cells = set()
+
+        # get a set containing all boundary cells
+        for cell in boundary_data.index_array:
+            fluid_cell = tuple([cell[coord] for coord in coords])
+            direction = self.stencil[cell['dir']]
+            boundary_cell = tuple([i + d for i, d in zip(fluid_cell, direction)])
+            boundary_cells.add(boundary_cell)
+
+        for cell in boundary_data.index_array:
+            fluid_cell = tuple([cell[coord] for coord in coords])
+            direction = self.stencil[cell['dir']]
+            ref_direction = direction
+            normal_direction = [0] * dim
+
+            for i in range(dim):
+                sub_direction = [0] * dim
+                sub_direction[i] = direction[i]
+                test_cell = tuple([x + y for x, y in zip(fluid_cell, sub_direction)])
+
+                if test_cell in boundary_cells:
+                    normal_direction[i] = direction[i]
+                    ref_direction = MirroredStencilDirections.mirror_stencil(ref_direction, i)
+
+            ref_direction = inverse_direction(ref_direction)
+            for i, cell_name in zip(range(dim), self.additional_data):
+                cell[cell_name[0]] = -normal_direction[i]
+            cell['ref_dir'] = self.stencil.index(ref_direction)
+
+    @property
+    def additional_data(self):
+        """Used internally only. For the FreeSlip boundary the information of the normal direction for each pdf
+        direction is needed. This information is stored in the index vector."""
+        if self.normal_direction:
+            return []
+        else:
+            data_type = create_type('int32')
+            wnz = [] if self.dim == 2 else [('wnz', data_type)]
+            data = [('wnx', data_type), ('wny', data_type)] + wnz
+            return data + [('ref_dir', data_type)]
+
+    @property
+    def additional_data_init_callback(self):
+        if self.normal_direction:
+            return None
+        else:
+            return self.init_callback
+
+    def get_additional_code_nodes(self, lb_method):
+        if self.normal_direction:
+            return [MirroredStencilDirections(self.stencil, self.mirror_axis)]
+        else:
+            return []
+
+    def __call__(self, f_out, f_in, dir_symbol, inv_dir, lb_method, index_field):
+        if self.normal_direction:
+            normal_direction = self.normal_direction
+            mirrored_stencil_symbol = MirroredStencilDirections._mirrored_symbol(self.mirror_axis)
+            mirrored_direction = inv_dir[sp.IndexedBase(mirrored_stencil_symbol, shape=(1,))[dir_symbol]]
+        else:
+            normal_direction = list()
+            for i, cell_name in zip(range(self.dim), self.additional_data):
+                normal_direction.append(index_field[0](cell_name[0]))
+            normal_direction = tuple(normal_direction)
+
+            mirrored_direction = index_field[0]('ref_dir')
+
+        return Assignment(f_in(inv_dir[dir_symbol]), f_in[normal_direction](mirrored_direction))
+
+
+# end class FreeSlip
+
 
 class UBB(LbBoundary):
     """Velocity bounce back boundary condition, enforcing specified velocity at obstacle
@@ -125,8 +245,7 @@ class UBB(LbBoundary):
         name: optional name of the boundary.
     """
 
-    def __init__(self, velocity, adapt_velocity_to_force=False, dim=None, name=None):
-        super(UBB, self).__init__(name)
+    def __init__(self, velocity, adapt_velocity_to_force=False, dim=None, name=None, data_type='double'):
         self._velocity = velocity
         self._adaptVelocityToForce = adapt_velocity_to_force
         if callable(self._velocity) and not dim:
@@ -134,13 +253,16 @@ class UBB(LbBoundary):
         elif not callable(self._velocity):
             dim = len(velocity)
         self.dim = dim
+        self.data_type = data_type
+
+        super(UBB, self).__init__(name)
 
     @property
     def additional_data(self):
         """ In case of the UBB boundary additional data is a velocity vector. This vector is added to each cell to
             realize velocity profiles for the inlet."""
         if self.velocity_is_callable:
-            return [('vel_%d' % (i,), create_type("double")) for i in range(self.dim)]
+            return [(f'vel_{i}', create_type(self.data_type)) for i in range(self.dim)]
         else:
             return []
 
@@ -172,12 +294,11 @@ class UBB(LbBoundary):
     def __call__(self, f_out, f_in, dir_symbol, inv_dir, lb_method, index_field):
         vel_from_idx_field = callable(self._velocity)
         vel = [index_field(f'vel_{i}') for i in range(self.dim)] if vel_from_idx_field else self._velocity
-        direction = dir_symbol
 
         assert self.dim == lb_method.dim, \
             f"Dimension of UBB ({self.dim}) does not match dimension of method ({lb_method.dim})"
 
-        neighbor_offset = NeighbourOffsetArrays.neighbour_offset(direction, lb_method.stencil)
+        neighbor_offset = NeighbourOffsetArrays.neighbour_offset(dir_symbol, lb_method.stencil)
 
         velocity = tuple(v_i.get_shifted(*neighbor_offset)
                          if isinstance(v_i, Field.Access) and not vel_from_idx_field
@@ -187,12 +308,13 @@ class UBB(LbBoundary):
         if self._adaptVelocityToForce:
             cqc = lb_method.conserved_quantity_computation
             shifted_vel_eqs = cqc.equilibrium_input_equations_from_init_values(velocity=velocity)
+            shifted_vel_eqs = shifted_vel_eqs.new_without_subexpressions()
             velocity = [eq.rhs for eq in shifted_vel_eqs.new_filtered(cqc.first_order_moment_symbols).main_assignments]
 
         c_s_sq = sp.Rational(1, 3)
         weight_of_direction = LbmWeightInfo.weight_of_direction
         vel_term = 2 / c_s_sq * sum([d_i * v_i for d_i, v_i in zip(neighbor_offset, velocity)]) * weight_of_direction(
-            direction, lb_method)
+            dir_symbol, lb_method)
 
         # Better alternative: in conserved value computation
         # rename what is currently called density to "virtual_density"
@@ -204,12 +326,12 @@ class UBB(LbBoundary):
             density_equations = cqc.output_equations_from_pdfs(pdf_field_accesses, {'density': density_symbol})
             density_symbol = lb_method.conserved_quantity_computation.defined_symbols()['density']
             result = density_equations.all_assignments
-            result += [Assignment(f_in(inv_dir[direction]),
-                                  f_out(direction) - vel_term * density_symbol)]
+            result += [Assignment(f_in(inv_dir[dir_symbol]),
+                                  f_out(dir_symbol) - vel_term * density_symbol)]
             return result
         else:
-            return [Assignment(f_in(inv_dir[direction]),
-                               f_out(direction) - vel_term)]
+            return [Assignment(f_in(inv_dir[dir_symbol]),
+                               f_out(dir_symbol) - vel_term)]
 
 
 # end class UBB
@@ -217,9 +339,9 @@ class UBB(LbBoundary):
 
 class SimpleExtrapolationOutflow(LbBoundary):
     r"""
-       Simple Outflow boundary condition :cite:`geier2015`, equation F.1 (listed below).
-       This boundary condition extrapolates missing populations from the last layer of
-       fluid cells onto the boundary by copying them in the normal direction.
+    Simple Outflow boundary condition :cite:`geier2015`, equation F.1 (listed below).
+    This boundary condition extrapolates missing populations from the last layer of
+    fluid cells onto the boundary by copying them in the normal direction.
 
     .. math ::
         f_{\overline{1}jkxyzt} = f_{\overline{1}jk(x - \Delta x)yzt}
@@ -258,16 +380,18 @@ class SimpleExtrapolationOutflow(LbBoundary):
         tangential_offset = tuple(offset - normal for offset, normal in zip(neighbor_offset, self.normal_direction))
 
         return Assignment(f_in.center(inv_dir[dir_symbol]), f_out[tangential_offset](inv_dir[dir_symbol]))
+
+
 # end class SimpleExtrapolationOutflow
 
 
 class ExtrapolationOutflow(LbBoundary):
     r"""
-        Outflow boundary condition :cite:`geier2015`, equation F.2, with u neglected (listed below).
-        This boundary condition interpolates populations missing on the boundary in normal direction. 
-        For this interpolation, the PDF values of the last time step are used. They are interpolated 
-        between fluid cell and boundary cell. To get the PDF values from the last time step an index 
-        array is used which stores them.
+    Outflow boundary condition :cite:`geier2015`, equation F.2, with u neglected (listed below).
+    This boundary condition interpolates populations missing on the boundary in normal direction.
+    For this interpolation, the PDF values of the last time step are used. They are interpolated
+    between fluid cell and boundary cell. To get the PDF values from the last time step an index
+    array is used which stores them.
 
     .. math ::
         f_{\overline{1}jkxyzt} = f_{\overline{1}jk(x - \Delta x)yz(t - \Delta t)} c \theta^{\frac{1}{2}}
@@ -291,11 +415,12 @@ class ExtrapolationOutflow(LbBoundary):
 
     def __init__(self, normal_direction, lb_method, dt=1, dx=1, name=None,
                  streaming_pattern='pull', zeroth_timestep=Timestep.BOTH,
-                 initial_density=None, initial_velocity=None):
+                 initial_density=None, initial_velocity=None, data_type='double'):
 
         self.lb_method = lb_method
         self.stencil = lb_method.stencil
         self.dim = len(self.stencil[0])
+
         if isinstance(normal_direction, str):
             normal_direction = direction_string_to_offset(normal_direction, dim=self.dim)
 
@@ -313,6 +438,8 @@ class ExtrapolationOutflow(LbBoundary):
         self.initial_velocity = initial_velocity
         self.equilibrium_calculation = None
 
+        self.data_type = data_type
+
         if initial_density and initial_velocity:
             equilibrium = lb_method.get_equilibrium(conserved_quantity_equations=AssignmentCollection([]))
             rho = lb_method.zeroth_order_equilibrium_moment_symbol
@@ -359,7 +486,7 @@ class ExtrapolationOutflow(LbBoundary):
     def additional_data(self):
         """Used internally only. For the ExtrapolationOutflow information of the previous PDF values is needed. This
         information is stored in the index vector."""
-        data = [('pdf', create_type("double")), ('pdf_nd', create_type("double"))]
+        data = [('pdf', create_type(self.data_type)), ('pdf_nd', create_type(self.data_type))]
         return data
 
     @property
@@ -417,8 +544,9 @@ class FixedDensity(LbBoundary):
     def __init__(self, density, name=None):
         if name is None:
             name = "Fixed Density " + str(density)
+        self.density = density
+
         super(FixedDensity, self).__init__(name)
-        self._density = density
 
     def __call__(self, f_out, f_in, dir_symbol, inv_dir, lb_method, index_field):
         def remove_asymmetric_part_of_main_assignments(assignment_collection, degrees_of_freedom):
@@ -438,7 +566,7 @@ class FixedDensity(LbBoundary):
 
         density_symbol = cqc.defined_symbols()['density']
 
-        density = self._density
+        density = self.density
         equilibrium_input = cqc.equilibrium_input_equations_from_init_values(density=density)
         equilibrium_input = equilibrium_input.new_without_subexpressions()
         density_eq = equilibrium_input.main_assignments[0]
@@ -459,6 +587,41 @@ class FixedDensity(LbBoundary):
 # end class FixedDensity
 
 
+class DiffusionDirichlet(LbBoundary):
+    """Boundary condition for advection-diffusion problems that fixes the concentration at the obstacle.
+
+    Args:
+        concentration: value of the concentration which should be set.
+        name: optional name of the boundary.
+    """
+
+    def __init__(self, concentration, name=None):
+        if name is None:
+            name = "Diffusion Dirichlet " + str(concentration)
+        self.concentration = concentration
+
+        super(DiffusionDirichlet, self).__init__(name)
+
+    def get_additional_code_nodes(self, lb_method):
+        """Return a list of code nodes that will be added in the generated code before the index field loop.
+
+        Args:
+            lb_method: Lattice Boltzmann method. See :func:`lbmpy.creationfunctions.create_lb_method`
+
+        Returns:
+            list containing LbmWeightInfo
+        """
+        return [LbmWeightInfo(lb_method)]
+
+    def __call__(self, f_out, f_in, dir_symbol, inv_dir, lb_method, index_field):
+        w_dir = LbmWeightInfo.weight_of_direction(dir_symbol, lb_method)
+        return [Assignment(f_in(inv_dir[dir_symbol]),
+                           2 * w_dir * self.concentration - f_out(dir_symbol))]
+
+
+# end class DiffusionDirichlet
+
+
 class NeumannByCopy(LbBoundary):
     """Neumann boundary condition which is implemented by coping the PDF values to achieve similar values at the fluid
        and the boundary node"""
@@ -479,13 +642,6 @@ class NeumannByCopy(LbBoundary):
         return [Assignment(f_in(inv_dir[dir_symbol]), f_out(inv_dir[dir_symbol])),
                 Assignment(f_out[neighbour_offset](dir_symbol), f_out(dir_symbol))]
 
-    def __hash__(self):
-        # All boundaries of these class behave equal -> should also be equal
-        return hash("NeumannByCopy")
-
-    def __eq__(self, other):
-        return type(other) == NeumannByCopy
-
 
 # end class NeumannByCopy
 
@@ -501,7 +657,7 @@ class StreamInConstant(LbBoundary):
 
     def __init__(self, constant, name=None):
         super(StreamInConstant, self).__init__(name)
-        self._constant = constant
+        self.constant = constant
 
     def get_additional_code_nodes(self, lb_method):
         """Return a list of code nodes that will be added in the generated code before the index field loop.
@@ -516,13 +672,7 @@ class StreamInConstant(LbBoundary):
 
     def __call__(self, f_out, f_in, dir_symbol, inv_dir, lb_method, index_field):
         neighbour_offset = NeighbourOffsetArrays.neighbour_offset(dir_symbol, lb_method.stencil)
-        return [Assignment(f_in(inv_dir[dir_symbol]), self._constant),
-                Assignment(f_out[neighbour_offset](dir_symbol), self._constant)]
+        return [Assignment(f_in(inv_dir[dir_symbol]), self.constant),
+                Assignment(f_out[neighbour_offset](dir_symbol), self.constant)]
 
-    def __hash__(self):
-        # All boundaries of these class behave equal -> should also be equal
-        return hash("StreamInConstant")
-
-    def __eq__(self, other):
-        return type(other) == StreamInConstant
 # end class StreamInConstant
